##############################################################################
#
#   Copyright (c) 2006 Zenoss, Inc. all rights reserved.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License.
#
###############################################################################

#
# contained below are functions that are called from different build
# environments.  they are intended to be reused between the rpm
# builder, rpath builder, and standalone install.sh/build.sh
#
# Christopher Blunck
#

test -r shared-functions.sh && . shared-functions.sh
test -r install-functions.sh && . install-functions.sh

# set up environment variables that will be used during the build
set_vars()
{
    # information fed to useradd if a shell account is created
    export OS_USERNAME=zenoss
    export OS_HOME=/home/${OS_USERNAME}
    export OS_SHELL=/bin/bash
    
    # the mysql events database name
    export MYSQLEVENTDB=events

    # try "root"
    export MYSQLUSER=zenoss

    # the mysql password for $MYSQLUSER
    export MYSQLPASS="zenoss"

    # the mysql server
    export MYSQLHOST="localhost"

    # the mysql root username and password
    export MYSQLROOTUSER="root"
    export MYSQLROOTPASSWD=""

    # the mysql port number
    export MYSQLPORT=3306
    
    # password for user "admin" in Zope
    export ZOPEPASSWORD="zenoss"

    # default to extracting the trunk
    if [ ! "$SVNTAG" ] ; then
        export SVNTAG="trunk"
    fi

    # default to enabling the setuid flags when necessary
    export SETUID="true"

    # the base directory for where zenoss will be installed
    export DESTDIR=/

    if [ -z "$ZENHOME" ] ; then
        export ZENHOME=`cat zenhome`
    fi

    if [ -z "$ZOPEHOME" ] ; then
        export ZOPEHOME=$ZENHOME/zopehome
    fi

    if [ -z "$VERBOSE" ] ; then
        VERBOSE=echo
    fi

    export MIGRATE="n"
    export NEW_INSTALL="y"
}


find_python()
{
    # set the PYTHON variable to one of the following values:
    #
    # "" : build our own python (almost always works)
    #
    # ${PLATFORM_PYTHON} : use the python executabled located at this location
    
    PYTHON=""
    if [ -x "${PLATFORM_PYTHON}" ]; then
        echo "Using platform python: ${PLATFORM_PYTHON}"
        PYTHON=${PLATFORM_PYTHON}

        # check the python to make sure it is compatible with our software
        $PYTHON -c 'import sys; sys.exit(sys.version_info[:2] != (2,6))' || fail python must be version 2.6
    fi

    export PYTHON
}


# set the MAKE and GNUMAKE environment variables
set_make()
{
    # set the MAKE variable
    export PYTHONPATH=$ZENHOME/lib/python
    export MAKE=make
    if gmake -f /dev/null -v >/dev/null 2>&1
    then
        export MAKE=gmake
        export GNUMAKE=gmake
    fi
}

# make sure we're not running as root.  if we are then fail
check_id()
{
    if [ -n "`id | grep uid=0`" ] ; then 
        fail "Do not install as root: make $ZENHOME writable for the zenoss user"
    fi
}

# remove the backup.tar if it exists
purge_backup()
{
    if [ -f backup.tar ] ; then
        rm backup.tar || fail Unable to remove a pre-existing backup file.
    fi
}

# create a backup if zenoss already is installed
check_preexist()
{
    if [ -d $ZENHOME/var ] ; then
        export NEW_INSTALL="n"
        echo "Stopping zenoss"
        $ZENHOME/bin/zenoss stop >/dev/null
        echo -n "Do you want me to keep your existing database? (Y/n) "
        read MIGRATE
        echo
        if [ "$MIGRATE" != "n" ]; then
            echo "Creating a backup of your existing data"
            ( cd $ZENHOME/var ; tar -cf - . ) > backup.tar || \
                fail Unable to create backup
            # double check
            test -s backup.tar || fail Unable to create backup
        fi
    fi
}

# handle ZENHOME symbolic links by preserving performance data
check_perf()
{
    # change old zenoss symlink to a real directory
    if [ -L $ZENHOME ] ; then
        REALDIR=`python -c 'import os; print os.path.realpath(os.getenv("ZENHOME"))'`
        rm -f $ZENHOME
        mkdir -p $ZENHOME
        mv $REALDIR/perf $ZENHOME
    fi
}

# prompt the user for mysql database settings
set_mysql()
{
    if [ "$MIGRATE" = "n" ]; then
        echo -n "MySQL server hostname [$MYSQLHOST]: "
        read response
        if [ ! -z "$response" ] ; then MYSQLHOST=$response ; fi

        echo -n "MySQL server root username [$MYSQLROOTUSER]: "
        read response
        if [ ! -z "$response" ] ; then MYSQLROOTUSER=$response ; fi

        echo -n "MySQL server root password [$MYSQLROOTPASSWD]: "
        noecho
        read response
        echook
        if [ ! -z "$response" ] ; then MYSQLROOTPASSWD=$response ; fi

        echo -n "MySQL event database name [$MYSQLEVENTDB]: "
        read response 
        if [ ! -z "$response" ] ; then MYSQLEVENTDB=$response ; fi

        echo -n "MySQL username for Zenoss $MYSQLEVENTDB database [$MYSQLUSER]: "
        read response
        if [ ! -z "$response" ] ; then MYSQLUSER=$response ; fi

        echo -n "MySQL password for $MYSQLUSER [$MYSQLPASS]: "
        noecho
        read response
        if [ ! -z "$response" ] ; then MYSQLPASS=$response ; fi
        echook

        echo -n "MySQL server port [$MYSQLPORT]: "
        read response
        if [ ! -z "$response" ] ; then MYSQLPORT=$response ; fi
    fi
}

# make sure all mysql environment variables are set and a devel version exists
check_mysql()
{
    if [ -z "$MYSQLHOST" ] ; then
        fail "You must set the MySQL hostname"
    fi

    if [ -z "$MYSQLUSER" ] ; then
        fail 'You must set the name of a MySQL user who has trigger and database create permissions (try "root")'
    fi
    
    if [ -z "$MYSQLPASS" ] ; then
        fail "You must set the password for $MYSQLUSER"
    fi
    
    # This is misplaced and should be moved elsewhere at some point
    if [ -z "$ZOPEPASSWORD" ] ; then
        fail 'You must set the password the user "admin" to access the Zenoss web pages within Zope.'
    fi

    MYSQLINC=`mysql_config --include | cut -c3-`
    if [ -z "$MYSQLINC" -o ! -d "$MYSQLINC" ] ; then
        fail 'You must have a MySQL development environment installed'
    fi
}

# check if ZENHOME is writable.  if it isn't then fail
check_writable()
{
    mkdir -p $ZENHOME 2>/dev/null
    touch $ZENHOME/.writable 2>/dev/null || fail "Unable to write to \$ZENHOME ($ZENHOME)"
}

# compare the current ZENHOME and the login ZENHOME.  fail if they differ
check_bashrc()
{
    LOGINZENHOME=`bash -i -c 'echo $ZENHOME'`
    if [ -n "$LOGINZENHOME" ] && [ "$LOGINZENHOME" != "$ZENHOME" ]; then
        fail $LOGINZENHOME from login shell is not the same as $ZENHOME
    fi
}

check_patch()
{
    patch=''
    for d  in /usr/local/bin /usr/bin ; do
       for p in gpatch patch ; do
          if [ -x $d/$p ] ; then
              patch="$d/$p"
          fi
       done
    done
    test -n "$patch" || fail unable to find the '"patch"' program
}

# compile the external libs
compile()
{
    if [ -z "$SKIPMAKE" ] ; then
        CC=`which cc`
        if [ ! -x "$CC" ]; then
            fail cc is required to build dependent libraries
        fi

        if [ "`uname`" != "SunOS" ]; then
            g++ -v 2>/dev/null || fail g++ is required to build dependent libraries
        fi
        
        ( $MAKE --version | grep -q GNU ) || fail $MAKE must be GNU make
        
        $MAKE \
            ZOPEPASSWORD="$ZOPEPASSWORD" \
            ZENHOME=$ZENHOME \
            DESTDIR=$DESTDIR \
            SETUID=$SETUID \
            SVNTAG=$SVNTAG \
                || fail "unable to build zenoss and prerequisites, see zenbuild.log"
    fi
}

# make sure the mysql-devel libraries are installed
check_mysql_config()
{
    # required to build MySQLdb
    mysql_config --version >/dev/null || fail mysql_config is not in the path
}

# Make sure build dependencies are installed
check_dependencies()
{
	which autoconf || fail autoconf is not in the path
	which swig || fail swig is not in the path
	which make || fail make is not in the path
	which svn || fail svn make is not in the path
}

# Make sure we can log into MySQL
check_mysql_login()
{
	if [ -z "${MYSQLROOTPASSWD}" ];then
		# The mysql root passwd var is null so avoid passing in -p option
		# otherwise -p by itself will cause mysql client to prompt for password when it
		# really doesn't need to. Was breaking CI builds.
		MYSQLROOTPASSWD_OPT=
	else
		MYSQLROOTPASSWD_OPT=-p$MYSQLROOTPASSWD
	fi
	mysql -h $MYSQLHOST -u$MYSQLROOTUSER $MYSQLROOTPASSWD_OPT -e 'SELECT VERSION()' || fail Could not connect to MySQL as root user
}

check_preupgrade_finished() {
    if [ "$NEW_INSTALL" = "n" ]; then
        OLDMAJOR=`cat $ZENHOME/Products/ZenModel/ZVersion.py | tail -n1 | cut -d\" -f2 | cut -d. -f1`
        if [ $OLDMAJOR = "2" ]; then
            if [ ! -f "${ZENHOME}/.zen3-preupgrade-finished" ]; then
                echo =========================================================
                echo Your system is not yet ready for upgrade. Please install
                echo ZenPacks.zenoss.PreUpgrade30 and let it run to completion before
                echo continuing the upgrade.
                echo
                echo If you have installed the ZenPack already, the preparation is not
                echo yet complete. Check the event console or $ZENHOME/log/zencatalog.log
                echo for progress. You can also ensure the upgrade script is still
                echo running with:
                echo
                echo    zencatalog restart
                echo =========================================================
                fail
            fi
        fi
    fi
}

# hurray we're done
get_happy()
{
    echo
    echo =========================================================
    echo zensocket must be setuid.  As root, execute the following:
    echo "chown root:`id -gn` $ZENHOME/bin/zensocket"
    echo "chmod 04750 $ZENHOME/bin/zensocket"
    echo =========================================================
    echo 
    echo Successfully installed Zenoss
    echo 
}

# performs safety checks to make sure the environment is good for install
run_checks() 
{
    # make sure we have at patch program
    check_patch

    # make sure we're not root
    check_id

    # handle situation where installation already exists
    check_preexist

    # Make sure the pre-upgrade script finished
    check_preupgrade_finished

    # check for running zenoss processes
    check_running

    # preserve performance data
    check_perf

    # make sure that variables are set and directories exist
    check_mysql

    # make sure that ZENHOME is writable
    check_writable

    # make sure ZENHOME is set in the user's login shell
    check_bashrc

    # make sure mysql development environment exists
    check_mysql_config

	# Make sure build dependencies are installed
	check_dependencies
}

# performs post-installation configuration to be performed after compile/install
configure()
{
    # create the events database
    create_database

    # load the events database with some initial data
    run_zeneventbuild

    # create a zope instance for zenoss
    run_mkzopeinstance

    # create the zeodb 
    run_mkzeoinstance

    # start up zeodb
    start_zeodb

    # wait for processes to start up
    sleep 2

    # populate the zeodb
    run_zenbuild

    # start the zope server
    start_zope
}

# sets up variables, runs checks, compiles, installs, and configures
build() 
{
    # initialize environment variables
    set_vars
    set_make

    # build the external libs and download zenoss
    compile
}

# sets up variables and configure
init()
{
    # initialize environment variables
    set_vars
    set_make

    # configure environment variables
    set_mysql

    # check mysql login
    check_mysql_login

    # perform post-installation configuration
    configure
}

